#ifndef MSP_GL_PROGRAM_H_
#define MSP_GL_PROGRAM_H_

#include <string>
#include <vector>
#include <msp/datafile/objectloader.h>
#include "bindable.h"
#include "datatype.h"
#include "gl.h"
#include "module.h"
#include "vertexformat.h"

namespace Msp {
namespace GL {

class Shader;

/**
A complete shader program.  Programs can be assembled of individual Shaders or
generated with a set of standard features.
*/
class Program: public Bindable<Program>
{
public:
	class Loader: public DataFile::CollectionObjectLoader<Program>
	{
	public:
		Loader(Program &, Collection &);

	private:
		virtual void finish();

		void attribute(unsigned, const std::string &);
		void fragment_shader(const std::string &);
		void geometry_shader(const std::string &);
		void module(const std::string &);
		void vertex_shader(const std::string &);
	};

private:
	class SpecializationLoader: public DataFile::Loader
	{
	private:
		std::map<std::string, int> &spec_values;

		static ActionMap shared_actions;

	public:
		SpecializationLoader(std::map<std::string, int> &);

	private:
		virtual void init_actions();

		void specialize_bool(const std::string &, bool);
		void specialize_int(const std::string &, int);
	};

public:
	typedef unsigned LayoutHash;
	struct UniformBlockInfo;

	struct UniformInfo
	{
		std::string name;
		const UniformBlockInfo *block;
		unsigned location;
		unsigned array_size;
		unsigned array_stride;
		unsigned matrix_stride;
		DataType type;

		UniformInfo();
	};

	struct UniformBlockInfo
	{
		std::string name;
		unsigned data_size;
		int bind_point;
		std::vector<const UniformInfo *> uniforms;
		LayoutHash layout_hash;

		UniformBlockInfo();
	};

	struct AttributeInfo
	{
		std::string name;
		unsigned location;
		unsigned array_size;
		DataType type;

		AttributeInfo();
	};

	typedef std::map<std::string, UniformInfo> UniformMap;
	typedef std::map<std::string, UniformBlockInfo> UniformBlockMap;
	typedef std::map<std::string, AttributeInfo> AttributeMap;

private:
	unsigned id;
	std::vector<unsigned> stage_ids;
	const Module *module;
	bool linked;
	UniformBlockMap uniform_blocks;
	UniformMap uniforms;
	LayoutHash uniform_layout_hash;
	AttributeMap attributes;

public:
	/// Constructs an empty Program with no shader stages attached.
	Program();

	/// Constructs a Program from unified source code using ProgramCompiler.
	DEPRECATED Program(const std::string &);

	/// Constructs a Program from vertex and fragment shader source code.
	DEPRECATED Program(const std::string &, const std::string &);

	/// Constructs a Program from a Module, with specialization constants.
	Program(const Module &, const std::map<std::string, int> & = std::map<std::string, int>());

private:
	void init();
public:
	virtual ~Program();

	void add_stages(const Module &, const std::map<std::string, int> & = std::map<std::string, int>());
private:
	unsigned add_stage(GLenum);
	void add_glsl_stages(const GlslModule &, const std::map<std::string, int> &);
	void compile_glsl_stage(unsigned);
	void add_spirv_stages(const SpirVModule &, const std::map<std::string, int> &);

public:
	DEPRECATED void attach_shader(Shader &shader);
	DEPRECATED void attach_shader_owned(Shader *shader);
	DEPRECATED void detach_shader(Shader &shader);
	DEPRECATED const std::vector<Shader *> &get_attached_shaders() const;

	DEPRECATED void bind_attribute(unsigned, const std::string &);
	DEPRECATED void bind_attribute(VertexAttribute, const std::string &);
	DEPRECATED void bind_fragment_data(unsigned, const std::string &);

	void link();
private:
	void query_uniforms();
	void query_uniform_blocks(const std::vector<UniformInfo *> &);
	void query_attributes();
	void collect_uniforms();
	void collect_block_uniforms(UniformBlockInfo &, const SpirVModule::Structure &, const std::string &, unsigned);
	void collect_attributes();
	void update_layout_hash();
	static LayoutHash compute_layout_hash(const std::vector<const UniformInfo *> &);
	static bool uniform_location_compare(const UniformInfo *, const UniformInfo *);
public:
	bool is_linked() const { return linked; }
	DEPRECATED std::string get_info_log() const;

	LayoutHash get_uniform_layout_hash() const { return uniform_layout_hash; }
	const UniformBlockMap &get_uniform_blocks() const { return uniform_blocks; }
	const UniformBlockInfo &get_uniform_block_info(const std::string &) const;
	const UniformMap &get_uniforms() const { return uniforms; }
	const UniformInfo &get_uniform_info(const std::string &) const;
	int get_uniform_location(const std::string &) const;
	const AttributeMap &get_attributes() const { return attributes; }
	const AttributeInfo &get_attribute_info(const std::string &) const;
	int get_attribute_location(const std::string &) const;

	void bind() const;
	static void unbind();
};

} // namespace GL
} // namespace Msp

#endif
