#ifndef MSP_GL_TEXTURECUBE_H_
#define MSP_GL_TEXTURECUBE_H_

#include <msp/gl/extensions/arb_texture_cube_map.h>
#include <msp/graphics/image.h>
#include "texture.h"
#include "vector.h"

namespace Msp {
namespace GL {

enum TextureCubeFace
{
	POSITIVE_X = GL_TEXTURE_CUBE_MAP_POSITIVE_X,
	NEGATIVE_X = GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	POSITIVE_Y = GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	NEGATIVE_Y = GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
	POSITIVE_Z = GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
	NEGATIVE_Z = GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
};

/**
Cube map texture, consisting of six square faces.  All of the faces must be of
the same size.  A cube map texture is addressed by three-dimensional texture
coordinates, with a principal range of [-1, 1].  The face is first selected
according to the largest coordinate, and the remaining two coordinates are used
to sample the face image.  The images are oriented so that the cross product of
the s and t axes will point into the cube.

All faces of a cube map texture must be allocated for it to be usable.

Requires OpenGL version 1.3.
*/
class TextureCube: public Texture
{
public:
	class Loader: public Msp::DataFile::DerivedObjectLoader<TextureCube, Texture::Loader>
	{
	public:
		Loader(TextureCube &);
		Loader(TextureCube &, Collection &);
	private:
		void init();

		void external_image(TextureCubeFace, const std::string &);
		void image_data(TextureCubeFace, const std::string &);
		void raw_data(TextureCubeFace, const std::string &);
		void storage(PixelFormat, unsigned);
		void storage_levels(PixelFormat, unsigned, unsigned);
	};

private:
	unsigned size;
	unsigned levels;
	/* Lowest six bits track allocation status of faces on the base level.  Bit
	seven is set if the entire base level is allocated. */
	unsigned allocated;

	static TextureCubeFace face_order[6];
	static Vector3 directions[6];
	static unsigned orientations[12];

public:
	TextureCube();

	/** Defines storage structure for the texture.  If lv is zero, the number
	of mipmap levels is automatically determined from storage dimensions.

	Must be called before an image can be uploaded.  Once storage is defined,
	it can't be changed. */
	void storage(PixelFormat fmt, unsigned size, unsigned lv = 0);

	DEPRECATED void storage(PixelComponents c, unsigned s, unsigned l = 0)
	{ storage(make_pixelformat(c, UNSIGNED_BYTE), s, l); }

	/** Allocates storage for the cube faces.  The contents are initially
	undefined.  If storage has already been allocated, does nothing. */
	void allocate(unsigned level);

	/** Updates the contents of a face.  Storage must be defined beforehand.
	The image data must have dimensions and format matching the defined
	storage. */
	void image(TextureCubeFace face, unsigned level, const void *data);

	DEPRECATED void image(TextureCubeFace face, unsigned level,
		PixelComponents comp, DataType type, const void *data);

	/** Updates a rectangular region of a face.  Storage must be defined
	beforehand.  The image data must be in a format mathing the defined storage
	and the update region must be fully inside the face. */
	void sub_image(TextureCubeFace face, unsigned level, int x, int y, unsigned w, unsigned h, const void *data);

	DEPRECATED void sub_image(TextureCubeFace face, unsigned level,
		int x, int y, unsigned w, unsigned h,
		PixelComponents comp, DataType type, const void *data);

	void image(TextureCubeFace, const Graphics::Image &);

	DEPRECATED void image(TextureCubeFace, const Graphics::Image &, bool);

	virtual void image(const Graphics::Image &, unsigned = 0);
	using Texture::image;

	unsigned get_size() const { return size; }
private:
	unsigned get_n_levels() const;
	unsigned get_level_size(unsigned) const;

public:
	/** Translates indices into face constants.  Valid indices are between 0
	and 5, inclusive. */
	static TextureCubeFace enumerate_faces(unsigned);

	static unsigned get_face_index(TextureCubeFace);

	/** Returns a vector pointing out of the face. */
	static const Vector3 &get_face_direction(TextureCubeFace);

	/** Returns a vector in the direction of the s axis of the face. */
	static const Vector3 &get_s_direction(TextureCubeFace);

	/** Returns a vector in the direction of the t axis of the face. */
	static const Vector3 &get_t_direction(TextureCubeFace);

	/** Returns a vector pointing to the center a texel. */
	Vector3 get_texel_direction(TextureCubeFace, unsigned, unsigned);

	virtual AsyncLoader *load(IO::Seekable &, const Resources * = 0) { return 0; }
	virtual UInt64 get_data_size() const;
	virtual void unload() { }
};

void operator>>(const LexicalConverter &, TextureCubeFace &);

} // namespace GL
} // namespace Msp

#endif
