#include <msp/core/algorithm.h>
#include <msp/core/raii.h>
#include <msp/strings/lexicalcast.h>
#include "finalize.h"
#include "glsl_error.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

PrecisionConverter::PrecisionConverter():
	stage(0)
{ }

void PrecisionConverter::apply(Stage &s)
{
	stage = &s;
	s.content.visit(*this);
	NodeRemover().apply(s, nodes_to_remove);
}

void PrecisionConverter::visit(Block &block)
{
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		if(&block==&stage->content)
			insert_point = i;
		(*i)->visit(*this);
	}
}

void PrecisionConverter::visit(Precision &prec)
{
	if(stage->required_features.gl_api==OPENGL_ES2)
		have_default.insert(prec.type);
	else
		nodes_to_remove.insert(&prec);
}

void PrecisionConverter::visit(VariableDeclaration &var)
{
	if(stage->required_features.gl_api!=OPENGL_ES2)
	{
		var.precision.clear();
		return;
	}

	const char *default_prec = (stage->type==Stage::FRAGMENT ? "mediump" : "highp");
	const TypeDeclaration *type = var.type_declaration;
	while(type)
	{
		if(dynamic_cast<const ImageTypeDeclaration *>(type))
		{
			default_prec = "lowp";
			break;
		}
		else if(const BasicTypeDeclaration *basic = dynamic_cast<const BasicTypeDeclaration *>(type))
		{
			if(basic->kind==BasicTypeDeclaration::INT || basic->kind==BasicTypeDeclaration::FLOAT)
				break;
			type = basic->base_type;
		}
		else
			return;
	}
	if(!type)
		return;

	if(!have_default.count(type->name))
	{
		Precision *prec = new Precision;
		prec->precision = default_prec;
		prec->type = type->name;
		stage->content.body.insert(insert_point, prec);

		have_default.insert(type->name);
	}
}


LegacyConverter::LegacyConverter():
	frag_out(0)
{ }

void LegacyConverter::apply(Stage &s, const Features &feat)
{
	stage = &s;
	features = feat;
	if(supports_stage(s.type))
	{
		s.content.visit(*this);
		NodeRemover().apply(s, nodes_to_remove);

		if(!stage->required_features.glsl_version)
			stage->required_features.glsl_version = Version(1, (stage->required_features.gl_api==OPENGL_ES2 ? 0 : 10));
	}
	else
		unsupported(format("Stage %s is not supported", Stage::get_stage_name(s.type)));
}

void LegacyConverter::unsupported(const string &reason)
{
	Diagnostic diagnostic;
	diagnostic.severity = Diagnostic::ERR;
	diagnostic.source = GENERATED_SOURCE;
	diagnostic.line = 0;
	diagnostic.message = reason;
	stage->diagnostics.push_back(diagnostic);
}

void LegacyConverter::visit(Block &block)
{
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		if(&block==&stage->content)
			uniform_insert_point = i;
		(*i)->visit(*this);
	}
}

bool LegacyConverter::check_version(const Version &feature_version) const
{
	if(features.glsl_version<feature_version)
		return false;
	else if(stage->required_features.glsl_version<feature_version)
		stage->required_features.glsl_version = feature_version;

	return true;
}

bool LegacyConverter::check_extension(bool Features::*extension) const
{
	if(!(features.*extension))
		return false;

	stage->required_features.*extension = true;

	return true;
}

bool LegacyConverter::supports_stage(Stage::Type st) const
{
	if(st==Stage::GEOMETRY)
	{
		if(features.gl_api==OPENGL_ES2)
			return check_version(Version(3, 20));
		else
			return check_version(Version(1, 50));
	}
	else
		return true;
}

bool LegacyConverter::supports_unified_interface_syntax() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 0));
	else
		return check_version(Version(1, 30));
}

void LegacyConverter::visit(VariableReference &var)
{
	if(var.declaration==frag_out && !supports_unified_interface_syntax())
	{
		var.name = "gl_FragColor";
		var.declaration = 0;
	}
}

void LegacyConverter::visit(Assignment &assign)
{
	TraversingVisitor::visit(assign);
	if(assign.target.declaration==frag_out && !supports_unified_interface_syntax())
		assign.target.declaration = 0;
}

bool LegacyConverter::supports_unified_sampling_functions() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 0));
	else
		return check_version(Version(1, 30));
}

void LegacyConverter::visit(FunctionCall &call)
{
	if(call.declaration && call.declaration->source==BUILTIN_SOURCE)
	{
		if(!call.name.compare(0, 7, "texture") && call.arguments.size()>=1)
		{
			const ImageTypeDeclaration *arg_image = dynamic_cast<const ImageTypeDeclaration *>(call.arguments.front()->type);
			if(arg_image && !supports_unified_sampling_functions())
			{
				string suffix = call.name.substr(7);
				call.name = (arg_image->shadow ? "shadow" : "texture");

				switch(arg_image->dimensions)
				{
				case ImageTypeDeclaration::ONE: call.name += "1D"; break;
				case ImageTypeDeclaration::TWO: call.name += "2D"; break;
				case ImageTypeDeclaration::THREE: call.name += "3D"; break;
				case ImageTypeDeclaration::CUBE: call.name += "Cube"; break;
				}

				if(arg_image->array)
				{
					/* Array textures and the unified sampling function name were
					both introduced in GLSL 1.30. */
					if(arg_image->dimensions==ImageTypeDeclaration::ONE || arg_image->dimensions==ImageTypeDeclaration::TWO)
						check_extension(&Features::ext_texture_array);
					call.name += "Array";
				}

				call.name += suffix;
			}
		}
	}

	TraversingVisitor::visit(call);
}

bool LegacyConverter::supports_interface_layouts() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 0));
	else if(check_version(Version(3, 30)))
		return true;
	else if(check_version(Version(1, 30)))
		return check_extension(&Features::arb_explicit_attrib_location);
	else
		return false;
}

bool LegacyConverter::supports_centroid_sampling() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 0));
	else if(check_version(Version(1, 20)))
		return true;
	else
		return check_extension(&Features::ext_gpu_shader4);
}

bool LegacyConverter::supports_sample_sampling() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 20));
	else if(check_version(Version(4, 0)))
		return true;
	else
		return check_extension(&Features::arb_gpu_shader5);
}

void LegacyConverter::visit(VariableDeclaration &var)
{
	if(var.layout)
	{
		for(vector<Layout::Qualifier>::const_iterator i=var.layout->qualifiers.begin(); i!=var.layout->qualifiers.end(); )
		{
			if(i->name=="location" && !supports_interface_layouts())
			{
				if(stage->type==Stage::VERTEX && var.interface=="in")
					stage->locations[var.name] = i->value;
				else if(stage->type==Stage::FRAGMENT && var.interface=="out")
				{
					if(check_extension(&Features::ext_gpu_shader4))
						stage->locations[var.name] = i->value;
					else if(i->value!=0)
						unsupported("EXT_gpu_shader4 required for multiple fragment shader outputs");
				}

				i = var.layout->qualifiers.erase(i);
			}
			else
				++i;
		}

		if(var.layout->qualifiers.empty())
			var.layout = 0;
	}

	if(var.sampling=="centroid")
	{
		if(!supports_centroid_sampling())
			var.sampling = string();
	}
	else if(var.sampling=="sample")
	{
		if(!supports_sample_sampling())
			var.sampling = string();
	}

	if((var.interface=="in" || var.interface=="out") && !supports_unified_interface_syntax())
	{
		if(stage->type==Stage::FRAGMENT && var.interface=="out")
		{
			frag_out = &var;
			nodes_to_remove.insert(&var);
		}
	}

	TraversingVisitor::visit(var);
}

bool LegacyConverter::supports_interface_blocks(const string &iface) const
{
	if(features.gl_api==OPENGL_ES2)
	{
		if(iface=="uniform")
			return check_version(Version(3, 0));
		else
			return check_version(Version(3, 20));
	}
	else if(check_version(Version(1, 50)))
		return true;
	else if(iface=="uniform")
		return check_extension(&Features::arb_uniform_buffer_object);
	else
		return false;
}

void LegacyConverter::visit(InterfaceBlock &iface)
{
	if(!supports_interface_blocks(iface.interface) && iface.type_declaration)
	{
		if(!iface.instance_name.empty())
			unsupported("ARB_uniform_buffer_object required for interface block instances");
		else if(iface.struct_declaration)
		{
			stage->content.body.splice(uniform_insert_point, iface.struct_declaration->members.body);
			nodes_to_remove.insert(&iface);
			nodes_to_remove.insert(iface.struct_declaration);
		}
		else
			/* If the interface block is an array, it should have an instance
			name too, so this should never be reached */
			throw logic_error("Unexpected interface block configuration");
	}
}

} // namespace SL
} // namespace GL
} // namespace Msp
