#include <msp/core/hash.h>
#include <msp/core/raii.h>
#include "generate.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

void ConstantIdAssigner::apply(Module &module, const Features &features)
{
	for(list<Stage>::iterator i=module.stages.begin(); i!=module.stages.end(); ++i)
		i->content.visit(*this);

	unsigned max_id = features.max_constant_id;
	for(vector<VariableDeclaration *>::iterator i=auto_constants.begin(); i!=auto_constants.end(); ++i)
	{
		unsigned id = hash32((*i)->name)%(max_id+1);
		while(used_ids.count(id))
			++id;

		vector<Layout::Qualifier> &qualifiers = (*i)->layout->qualifiers;
		for(vector<Layout::Qualifier>::iterator j=qualifiers.begin(); j!=qualifiers.end(); ++j)
			if(j->name=="constant_id")
			{
				j->value = id;
				break;
			}

		used_ids.insert(id);
	}
}

void ConstantIdAssigner::visit(VariableDeclaration &var)
{
	if(var.layout)
	{
		vector<Layout::Qualifier> &qualifiers = var.layout->qualifiers;
		for(vector<Layout::Qualifier>::iterator i=qualifiers.begin(); i!=qualifiers.end(); ++i)
			if(i->name=="constant_id" && i->has_value)
			{
				if(i->value==-1)
					auto_constants.push_back(&var);
				else
					used_ids.insert(i->value);
				break;
			}
	}
}


InterfaceGenerator::InterfaceGenerator():
	stage(0),
	function_scope(false),
	copy_block(false),
	iface_target_block(0)
{ }

string InterfaceGenerator::get_out_prefix(Stage::Type type)
{
	if(type==Stage::VERTEX)
		return "_vs_out_";
	else if(type==Stage::GEOMETRY)
		return "_gs_out_";
	else
		return string();
}

void InterfaceGenerator::apply(Stage &s)
{
	stage = &s;
	iface_target_block = &stage->content;
	if(stage->previous)
		in_prefix = get_out_prefix(stage->previous->type);
	out_prefix = get_out_prefix(stage->type);
	s.content.visit(*this);
	NodeRemover().apply(s, nodes_to_remove);
}

void InterfaceGenerator::visit(Block &block)
{
	SetForScope<Block *> set_block(current_block, &block);
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		assignment_insert_point = i;
		if(&block==&stage->content)
			iface_insert_point = i;

		(*i)->visit(*this);
	}
}

string InterfaceGenerator::change_prefix(const string &name, const string &prefix) const
{
	unsigned offset = (name.compare(0, in_prefix.size(), in_prefix) ? 0 : in_prefix.size());
	return prefix+name.substr(offset);
}

VariableDeclaration *InterfaceGenerator::generate_interface(VariableDeclaration &var, const string &iface, const string &name)
{
	if(stage->content.variables.count(name))
		return 0;

	if(stage->type==Stage::GEOMETRY && !copy_block && var.interface=="out" && var.array)
		return 0;

	VariableDeclaration* iface_var = new VariableDeclaration;
	iface_var->sampling = var.sampling;
	iface_var->interface = iface;
	iface_var->type = var.type;
	iface_var->name = name;
	/* Geometry shader inputs are always arrays.  But if we're bringing in an
	entire block, the array is on the block and not individual variables. */
	if(stage->type==Stage::GEOMETRY && !copy_block)
		iface_var->array = ((var.array && var.interface!="in") || iface=="in");
	else
		iface_var->array = var.array;
	if(iface_var->array)
		iface_var->array_size = var.array_size;
	if(iface=="in")
	{
		iface_var->layout = var.layout;
		iface_var->linked_declaration = &var;
		var.linked_declaration = iface_var;
	}

	iface_target_block->body.insert(iface_insert_point, iface_var);
	iface_target_block->variables.insert(make_pair(name, iface_var));
	if(iface_target_block==&stage->content && iface=="in")
		declared_inputs.push_back(iface_var);

	return iface_var;
}

InterfaceBlock *InterfaceGenerator::generate_interface(InterfaceBlock &out_block)
{
	if(stage->interface_blocks.count("in"+out_block.block_name))
		return 0;

	InterfaceBlock *in_block = new InterfaceBlock;
	in_block->interface = "in";
	in_block->block_name = out_block.block_name;
	in_block->members = new Block;
	in_block->instance_name = out_block.instance_name;
	if(stage->type==Stage::GEOMETRY)
		in_block->array = true;
	else
		in_block->array = out_block.array;
	in_block->linked_block = &out_block;
	out_block.linked_block = in_block;

	{
		SetFlag set_copy(copy_block, true);
		SetForScope<Block *> set_target(iface_target_block, in_block->members.get());
		SetForScope<NodeList<Statement>::iterator> set_ins_pt(iface_insert_point, in_block->members->body.end());
		if(out_block.struct_declaration)
			out_block.struct_declaration->members.visit(*this);
		else if(out_block.members)
			out_block.members->visit(*this);
	}

	iface_target_block->body.insert(iface_insert_point, in_block);
	stage->interface_blocks.insert(make_pair("in"+in_block->block_name, in_block));
	if(!in_block->instance_name.empty())
		stage->interface_blocks.insert(make_pair("_"+in_block->instance_name, in_block));

	SetFlag set_scope(function_scope, false);
	SetForScope<Block *> set_block(current_block, &stage->content);
	in_block->visit(*this);

	return in_block;
}

ExpressionStatement &InterfaceGenerator::insert_assignment(const string &left, Expression *right)
{
	Assignment *assign = new Assignment;
	VariableReference *ref = new VariableReference;
	ref->name = left;
	assign->left = ref;
	assign->oper = &Operator::get_operator("=", Operator::BINARY);
	assign->right = right;

	ExpressionStatement *stmt = new ExpressionStatement;
	stmt->expression = assign;
	current_block->body.insert(assignment_insert_point, stmt);
	stmt->visit(*this);

	return *stmt;
}

void InterfaceGenerator::visit(VariableReference &var)
{
	if(var.declaration || !stage->previous)
		return;
	/* Don't pull a variable from previous stage if we just generated an output
	interface in this stage */
	if(stage->content.variables.count(var.name))
		return;

	const map<string, VariableDeclaration *> &prev_vars = stage->previous->content.variables;
	map<string, VariableDeclaration *>::const_iterator i = prev_vars.find(var.name);
	if(i==prev_vars.end() || i->second->interface!="out")
		i = prev_vars.find(in_prefix+var.name);
	if(i!=prev_vars.end() && i->second->interface=="out")
	{
		if(stage->type==Stage::GEOMETRY && i->second->array)
			stage->diagnostics.push_back(Diagnostic(Diagnostic::WARN, var.source, var.line,
				format("Can't access '%s' through automatic interface because it's an array", var.name)));
		else
		{
			generate_interface(*i->second, "in", i->second->name);
			var.name = i->second->name;
		}
		return;
	}

	const map<string, InterfaceBlock *> &prev_blocks = stage->previous->interface_blocks;
	map<string, InterfaceBlock *>::const_iterator j = prev_blocks.find("_"+var.name);
	if(j!=prev_blocks.end() && j->second->interface=="out")
	{
		generate_interface(*j->second);
		/* Let VariableResolver convert the variable reference into an interface
		block reference. */
		return;
	}

	for(j=prev_blocks.begin(); j!=prev_blocks.end(); ++j)
		if(j->second->instance_name.empty() && j->second->struct_declaration)
		{
			const map<string, VariableDeclaration *> &iface_vars = j->second->struct_declaration->members.variables;
			i = iface_vars.find(var.name);
			if(i!=iface_vars.end())
			{
				generate_interface(*j->second);
				return;
			}
		}
}

void InterfaceGenerator::visit(VariableDeclaration &var)
{
	if(copy_block)
		generate_interface(var, "in", var.name);
	else if(var.interface=="out")
	{
		/* For output variables in function scope, generate a global interface
		and replace the local declaration with an assignment. */
		VariableDeclaration *out_var = 0;
		if(function_scope && (out_var=generate_interface(var, "out", var.name)))
		{
			out_var->source = var.source;
			out_var->line = var.line;
			nodes_to_remove.insert(&var);
			if(var.init_expression)
			{
				ExpressionStatement &stmt = insert_assignment(var.name, var.init_expression->clone());
				stmt.source = var.source;
				stmt.line = var.line;
				return;
			}
		}
	}
	else if(var.interface=="in" && current_block==&stage->content)
	{
		if(var.name.compare(0, 3, "gl_"))
			declared_inputs.push_back(&var);

		/* Try to link input variables in global scope with output variables from
		previous stage. */
		if(!var.linked_declaration && stage->previous)
		{
			const map<string, VariableDeclaration *> &prev_vars = stage->previous->content.variables;
			map<string, VariableDeclaration *>::const_iterator i = prev_vars.find(var.name);
			if(i!=prev_vars.end() && i->second->interface=="out")
			{
				var.linked_declaration = i->second;
				i->second->linked_declaration = &var;
			}
		}
	}

	TraversingVisitor::visit(var);
}

void InterfaceGenerator::visit(InterfaceBlock &iface)
{
	if(iface.interface=="in")
	{
		/* Try to link input blocks with output blocks sharing the same block
		name from previous stage. */
		if(!iface.linked_block && stage->previous)
		{
			const map<string, InterfaceBlock *> &prev_blocks = stage->previous->interface_blocks;
			map<string, InterfaceBlock *>::const_iterator i = prev_blocks.find("out"+iface.block_name);
			if(i!=prev_blocks.end())
			{
				iface.linked_block = i->second;
				i->second->linked_block = &iface;
			}
		}
	}

	TraversingVisitor::visit(iface);
}

void InterfaceGenerator::visit(FunctionDeclaration &func)
{
	SetFlag set_scope(function_scope, true);
	// Skip parameters because they're not useful here
	func.body.visit(*this);
}

void InterfaceGenerator::visit(Passthrough &pass)
{
	// Pass through all input variables declared so far.
	vector<VariableDeclaration *> pass_vars = declared_inputs;

	if(stage->previous)
	{
		const map<string, VariableDeclaration *> &prev_vars = stage->previous->content.variables;
		for(map<string, VariableDeclaration *>::const_iterator i=prev_vars.begin(); i!=prev_vars.end(); ++i)
		{
			if(i->second->interface!="out")
				continue;

			/* Pass through output variables from the previous stage, but only
			those which are not already linked to an input here. */
			if(!i->second->linked_declaration && generate_interface(*i->second, "in", i->second->name))
				pass_vars.push_back(i->second);
		}
	}

	if(stage->type==Stage::GEOMETRY)
	{
		/* Special case for geometry shader: copy gl_Position from input to
		output. */
		InterfaceBlockReference *ref = new InterfaceBlockReference;
		ref->name = "gl_in";

		BinaryExpression *subscript = new BinaryExpression;
		subscript->left = ref;
		subscript->oper = &Operator::get_operator("[", Operator::BINARY);
		subscript->right = pass.subscript;

		MemberAccess *memacc = new MemberAccess;
		memacc->left = subscript;
		memacc->member = "gl_Position";

		insert_assignment("gl_Position", memacc);
	}

	for(vector<VariableDeclaration *>::const_iterator i=pass_vars.begin(); i!=pass_vars.end(); ++i)
	{
		string out_name = change_prefix((*i)->name, out_prefix);
		generate_interface(**i, "out", out_name);

		VariableReference *ref = new VariableReference;
		ref->name = (*i)->name;
		if(pass.subscript)
		{
			BinaryExpression *subscript = new BinaryExpression;
			subscript->left = ref;
			subscript->oper = &Operator::get_operator("[", Operator::BINARY);
			subscript->right = pass.subscript;
			insert_assignment(out_name, subscript);
		}
		else
			insert_assignment(out_name, ref);
	}

	nodes_to_remove.insert(&pass);
}

} // namespace SL
} // namespace GL
} // namespace Msp
