#include <msp/core/raii.h>
#include <msp/strings/format.h>
#include "optimize.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

ConstantSpecializer::ConstantSpecializer():
	values(0)
{ }

void ConstantSpecializer::apply(Stage &stage, const map<string, int> &v)
{
	values = &v;
	stage.content.visit(*this);
}

void ConstantSpecializer::visit(VariableDeclaration &var)
{
	bool specializable = false;
	if(var.layout)
	{
		vector<Layout::Qualifier> &qualifiers = var.layout->qualifiers;
		for(vector<Layout::Qualifier>::iterator i=qualifiers.begin(); (!specializable && i!=qualifiers.end()); ++i)
			if(i->name=="constant_id")
			{
				specializable = true;
				qualifiers.erase(i);
			}

		if(qualifiers.empty())
			var.layout = 0;
	}

	if(specializable)
	{
		map<string, int>::const_iterator i = values->find(var.name);
		if(i!=values->end())
		{
			RefPtr<Literal> literal = new Literal;
			if(var.type=="bool")
			{
				literal->token = (i->second ? "true" : "false");
				literal->value = static_cast<bool>(i->second);
			}
			else if(var.type=="int")
			{
				literal->token = lexical_cast<string>(i->second);
				literal->value = i->second;
			}
			var.init_expression = literal;
		}
	}
}


InlineableFunctionLocator::InlineableFunctionLocator():
	current_function(0),
	return_count(0)
{ }

void InlineableFunctionLocator::visit(FunctionCall &call)
{
	FunctionDeclaration *def = call.declaration;
	if(def)
		def = def->definition;

	if(def)
	{
		unsigned &count = refcounts[def];
		++count;
		/* Don't inline functions which are called more than once or are called
		recursively. */
		if(count>1 || def==current_function)
			inlineable.erase(def);
	}

	TraversingVisitor::visit(call);
}

void InlineableFunctionLocator::visit(FunctionDeclaration &func)
{
	bool has_out_params = false;
	for(NodeArray<VariableDeclaration>::const_iterator i=func.parameters.begin(); (!has_out_params && i!=func.parameters.end()); ++i)
		has_out_params = ((*i)->interface=="out");

	unsigned &count = refcounts[func.definition];
	if(count<=1 && !has_out_params)
		inlineable.insert(func.definition);

	SetForScope<FunctionDeclaration *> set(current_function, &func);
	return_count = 0;
	TraversingVisitor::visit(func);
}

void InlineableFunctionLocator::visit(Conditional &cond)
{
	TraversingVisitor::visit(cond);
	inlineable.erase(current_function);
}

void InlineableFunctionLocator::visit(Iteration &iter)
{
	TraversingVisitor::visit(iter);
	inlineable.erase(current_function);
}

void InlineableFunctionLocator::visit(Return &ret)
{
	TraversingVisitor::visit(ret);
	if(return_count)
		inlineable.erase(current_function);
	++return_count;
}


InlineContentInjector::InlineContentInjector():
	source_func(0),
	pass(DEPENDS)
{ }

const string &InlineContentInjector::apply(Stage &stage, FunctionDeclaration &target_func, Block &tgt_blk, const NodeList<Statement>::iterator &ins_pt, FunctionCall &call)
{
	source_func = call.declaration->definition;

	// Collect all declarations the inlined function depends on.
	pass = DEPENDS;
	source_func->visit(*this);

	/* Populate referenced_names from the target function so we can rename
	variables from the inlined function that would conflict. */
	pass = REFERENCED;
	target_func.visit(*this);

	/* Inline and rename passes must be interleaved so used variable names are
	known when inlining the return statement. */
	pass = INLINE;
	staging_block.parent = &tgt_blk;
	staging_block.variables.clear();

	std::vector<RefPtr<VariableDeclaration> > params;
	params.reserve(source_func->parameters.size());
	for(NodeArray<VariableDeclaration>::iterator i=source_func->parameters.begin(); i!=source_func->parameters.end(); ++i)
	{
		RefPtr<VariableDeclaration> var = (*i)->clone();
		var->interface.clear();

		SetForScope<Pass> set_pass(pass, RENAME);
		var->visit(*this);

		staging_block.body.push_back_nocopy(var);
		params.push_back(var);
	}

	for(NodeList<Statement>::iterator i=source_func->body.body.begin(); i!=source_func->body.body.end(); ++i)
	{
		r_inlined_statement = 0;
		(*i)->visit(*this);
		if(!r_inlined_statement)
			r_inlined_statement = (*i)->clone();

		SetForScope<Pass> set_pass(pass, RENAME);
		r_inlined_statement->visit(*this);

		staging_block.body.push_back_nocopy(r_inlined_statement);
	}

	/* Now collect names from the staging block.  Local variables that would
	have conflicted with the target function were renamed earlier. */
	pass = REFERENCED;
	referenced_names.clear();
	staging_block.variables.clear();
	staging_block.visit(*this);

	/* Rename variables in the target function so they don't interfere with
	global identifiers used by the source function. */
	pass = RENAME;
	staging_block.parent = source_func->body.parent;
	target_func.visit(*this);

	// Put the argument expressions in place after all renaming has been done.
	for(unsigned i=0; i<source_func->parameters.size(); ++i)
		params[i]->init_expression = call.arguments[i]->clone();

	tgt_blk.body.splice(ins_pt, staging_block.body);

	NodeReorderer().apply(stage, target_func, dependencies);

	return r_result_name;
}

void InlineContentInjector::visit(VariableReference &var)
{
	if(pass==RENAME)
	{
		map<string, VariableDeclaration *>::const_iterator i = staging_block.variables.find(var.name);
		if(i!=staging_block.variables.end())
			var.name = i->second->name;
	}
	else if(pass==DEPENDS && var.declaration)
	{
		dependencies.insert(var.declaration);
		var.declaration->visit(*this);
	}
	else if(pass==REFERENCED)
		referenced_names.insert(var.name);
}

void InlineContentInjector::visit(InterfaceBlockReference &iface)
{
	if(pass==DEPENDS && iface.declaration)
	{
		dependencies.insert(iface.declaration);
		iface.declaration->visit(*this);
	}
	else if(pass==REFERENCED)
		referenced_names.insert(iface.name);
}

void InlineContentInjector::visit(FunctionCall &call)
{
	if(pass==DEPENDS && call.declaration)
		dependencies.insert(call.declaration);
	else if(pass==REFERENCED)
		referenced_names.insert(call.name);
	TraversingVisitor::visit(call);
}

void InlineContentInjector::visit(VariableDeclaration &var)
{
	TraversingVisitor::visit(var);

	if(pass==RENAME)
	{
		/* Check against conflicts with the other context as well as variables
		already renamed here. */
		bool conflict = (staging_block.variables.count(var.name) || referenced_names.count(var.name));
		staging_block.variables[var.name] = &var;
		if(conflict)
		{
			string mapped_name = get_unused_variable_name(staging_block, var.name);
			if(mapped_name!=var.name)
			{
				staging_block.variables[mapped_name] = &var;
				var.name = mapped_name;
			}
		}
	}
	else if(pass==DEPENDS && var.type_declaration)
	{
		dependencies.insert(var.type_declaration);
		var.type_declaration->visit(*this);
	}
	else if(pass==REFERENCED)
		referenced_names.insert(var.type);
}

void InlineContentInjector::visit(Return &ret)
{
	TraversingVisitor::visit(ret);

	if(pass==INLINE && ret.expression)
	{
		// Create a new variable to hold the return value of the inlined function.
		r_result_name = get_unused_variable_name(staging_block, "_return");
		RefPtr<VariableDeclaration> var = new VariableDeclaration;
		var->source = ret.source;
		var->line = ret.line;
		var->type = source_func->return_type;
		var->name = r_result_name;
		var->init_expression = ret.expression->clone();
		r_inlined_statement = var;
	}
}


FunctionInliner::FunctionInliner():
	current_function(0),
	r_any_inlined(false),
	r_inlined_here(false)
{ }

bool FunctionInliner::apply(Stage &s)
{
	stage = &s;
	inlineable = InlineableFunctionLocator().apply(s);
	r_any_inlined = false;
	s.content.visit(*this);
	return r_any_inlined;
}

void FunctionInliner::visit(RefPtr<Expression> &ptr)
{
	r_inline_result = 0;
	ptr->visit(*this);
	if(r_inline_result)
	{
		ptr = r_inline_result;
		r_any_inlined = true;
	}
	r_inline_result = 0;
}

void FunctionInliner::visit(Block &block)
{
	SetForScope<Block *> set_block(current_block, &block);
	SetForScope<NodeList<Statement>::iterator> save_insert_point(insert_point, block.body.begin());
	for(NodeList<Statement>::iterator i=block.body.begin(); (!r_inlined_here && i!=block.body.end()); ++i)
	{
		insert_point = i;
		(*i)->visit(*this);
	}
}

void FunctionInliner::visit(FunctionCall &call)
{
	for(NodeArray<Expression>::iterator i=call.arguments.begin(); (!r_inlined_here && i!=call.arguments.end()); ++i)
		visit(*i);

	if(r_inlined_here)
		return;

	FunctionDeclaration *def = call.declaration;
	if(def)
		def = def->definition;

	if(def && inlineable.count(def))
	{
		string result_name = InlineContentInjector().apply(*stage, *current_function, *current_block, insert_point, call);

		// This will later get removed by UnusedVariableRemover.
		if(result_name.empty())
			result_name = "_msp_unused_from_inline";

		RefPtr<VariableReference> ref = new VariableReference;
		ref->name = result_name;
		r_inline_result = ref;

		/* Inlined variables need to be resolved before this function can be
		inlined further. */
		inlineable.erase(current_function);
		r_inlined_here = true;
	}
}

void FunctionInliner::visit(FunctionDeclaration &func)
{
	SetForScope<FunctionDeclaration *> set_func(current_function, &func);
	TraversingVisitor::visit(func);
	r_inlined_here = false;
}

void FunctionInliner::visit(Iteration &iter)
{
	/* Visit the initialization statement before entering the loop body so the
	inlined statements get inserted outside. */
	if(iter.init_statement)
		iter.init_statement->visit(*this);

	SetForScope<Block *> set_block(current_block, &iter.body);
	/* Skip the condition and loop expression parts because they're not properly
	inside the body block.  Inlining anything into them will require a more
	comprehensive transformation. */
	iter.body.visit(*this);
}


ExpressionInliner::ExpressionInfo::ExpressionInfo():
	expression(0),
	assign_scope(0),
	inline_point(0),
	trivial(false),
	available(true)
{ }


ExpressionInliner::ExpressionInliner():
	r_ref_info(0),
	r_any_inlined(false),
	r_trivial(false),
	mutating(false),
	iteration_init(false),
	iteration_body(0),
	r_oper(0)
{ }

bool ExpressionInliner::apply(Stage &s)
{
	s.content.visit(*this);
	return r_any_inlined;
}

void ExpressionInliner::inline_expression(Expression &expr, RefPtr<Expression> &ptr)
{
	ptr = expr.clone();
	r_any_inlined = true;
}

void ExpressionInliner::visit(Block &block)
{
	TraversingVisitor::visit(block);

	for(map<string, VariableDeclaration *>::iterator i=block.variables.begin(); i!=block.variables.end(); ++i)
	{
		map<Assignment::Target, ExpressionInfo>::iterator j = expressions.lower_bound(i->second);
		for(; (j!=expressions.end() && j->first.declaration==i->second); )
		{
			if(j->second.expression && j->second.inline_point)
				inline_expression(*j->second.expression, *j->second.inline_point);

			expressions.erase(j++);
		}
	}

	/* Expressions assigned in this block may depend on local variables of the
	block.  If this is a conditionally executed block, the assignments might not
	always happen.  Mark the expressions as not available to any outer blocks. */
	for(map<Assignment::Target, ExpressionInfo>::iterator i=expressions.begin(); i!=expressions.end(); ++i)
		if(i->second.assign_scope==&block)
			i->second.available = false;
}

void ExpressionInliner::visit(RefPtr<Expression> &expr)
{
	r_ref_info = 0;
	expr->visit(*this);
	if(r_ref_info && r_ref_info->expression && r_ref_info->available)
	{
		if(iteration_body && !r_ref_info->trivial)
		{
			/* Don't inline non-trivial expressions which were assigned outside
			an iteration statement.  The iteration may run multiple times, which
			would cause the expression to also be evaluated multiple times. */
			Block *i = r_ref_info->assign_scope;
			for(; (i && i!=iteration_body); i=i->parent) ;
			if(!i)
				return;
		}

		if(r_ref_info->trivial)
			inline_expression(*r_ref_info->expression, expr);
		else
			/* Record the inline point for a non-trivial expression but don't
			inline it yet.  It might turn out it shouldn't be inlined after all. */
			r_ref_info->inline_point = &expr;
	}
	r_oper = expr->oper;
	r_ref_info = 0;
}

void ExpressionInliner::visit(VariableReference &var)
{
	if(var.declaration)
	{
		map<Assignment::Target, ExpressionInfo>::iterator i = expressions.find(var.declaration);
		if(i!=expressions.end())
		{
			/* If a non-trivial expression is referenced multiple times, don't
			inline it. */
			if(i->second.inline_point && !i->second.trivial)
				i->second.expression = 0;
			/* Mutating expressions are analogous to self-referencing assignments
			and prevent inlining. */
			if(mutating)
				i->second.expression = 0;
			r_ref_info = &i->second;
		}
	}
}

void ExpressionInliner::visit(MemberAccess &memacc)
{
	visit(memacc.left);
	r_trivial = false;
}

void ExpressionInliner::visit(Swizzle &swizzle)
{
	visit(swizzle.left);
	r_trivial = false;
}

void ExpressionInliner::visit(UnaryExpression &unary)
{
	SetFlag set_target(mutating, mutating || unary.oper->token[1]=='+' || unary.oper->token[1]=='-');
	visit(unary.expression);
	r_trivial = false;
}

void ExpressionInliner::visit(BinaryExpression &binary)
{
	visit(binary.left);
	{
		SetFlag clear_target(mutating, false);
		visit(binary.right);
	}
	r_trivial = false;
}

void ExpressionInliner::visit(Assignment &assign)
{
	{
		SetFlag set_target(mutating);
		visit(assign.left);
	}
	r_oper = 0;
	visit(assign.right);

	map<Assignment::Target, ExpressionInfo>::iterator i = expressions.find(assign.target);
	if(i!=expressions.end())
	{
		/* Self-referencing assignments can't be inlined without additional
		work.  Just clear any previous expression. */
		i->second.expression = (assign.self_referencing ? 0 : assign.right.get());
		i->second.assign_scope = current_block;
		i->second.inline_point = 0;
		i->second.available = true;
	}

	r_trivial = false;
}

void ExpressionInliner::visit(TernaryExpression &ternary)
{
	visit(ternary.condition);
	visit(ternary.true_expr);
	visit(ternary.false_expr);
	r_trivial = false;
}

void ExpressionInliner::visit(FunctionCall &call)
{
	TraversingVisitor::visit(call);
	r_trivial = false;
}

void ExpressionInliner::visit(VariableDeclaration &var)
{
	r_oper = 0;
	r_trivial = true;
	TraversingVisitor::visit(var);

	bool constant = var.constant;
	if(constant && var.layout)
	{
		for(vector<Layout::Qualifier>::const_iterator i=var.layout->qualifiers.begin(); (constant && i!=var.layout->qualifiers.end()); ++i)
			constant = (i->name!="constant_id");
	}

	/* Only inline global variables if they're constant and have trivial
	initializers.  Non-constant variables could change in ways which are hard to
	analyze and non-trivial expressions could be expensive to inline.  */
	if((current_block->parent || (constant && r_trivial)) && var.interface.empty())
	{
		ExpressionInfo &info = expressions[&var];
		/* Assume variables declared in an iteration initialization statement
		will have their values change throughout the iteration. */
		info.expression = (iteration_init ? 0 : var.init_expression.get());
		info.assign_scope = current_block;
		info.trivial = r_trivial;
	}
}

void ExpressionInliner::visit(Iteration &iter)
{
	SetForScope<Block *> set_block(current_block, &iter.body);
	if(iter.init_statement)
	{
		SetFlag set_init(iteration_init);
		iter.init_statement->visit(*this);
	}

	SetForScope<Block *> set_body(iteration_body, &iter.body);
	if(iter.condition)
		visit(iter.condition);
	iter.body.visit(*this);
	if(iter.loop_expression)
		visit(iter.loop_expression);
}


BasicTypeDeclaration::Kind ConstantFolder::get_value_kind(const Variant &value)
{
	if(value.check_type<bool>())
		return BasicTypeDeclaration::BOOL;
	else if(value.check_type<int>())
		return BasicTypeDeclaration::INT;
	else if(value.check_type<float>())
		return BasicTypeDeclaration::FLOAT;
	else
		return BasicTypeDeclaration::VOID;
}

template<typename T>
T ConstantFolder::evaluate_logical(char oper, T left, T right)
{
	switch(oper)
	{
	case '&': return left&right;
	case '|': return left|right;
	case '^': return left^right;
	default: return T();
	}
}

template<typename T>
bool ConstantFolder::evaluate_relation(const char *oper, T left, T right)
{
	switch(oper[0]|oper[1])
	{
	case '<': return left<right;
	case '<'|'=': return left<=right;
	case '>': return left>right;
	case '>'|'=': return left>=right;
	default: return false;
	}
}

template<typename T>
T ConstantFolder::evaluate_arithmetic(char oper, T left, T right)
{
	switch(oper)
	{
	case '+': return left+right;
	case '-': return left-right;
	case '*': return left*right;
	case '/': return left/right;
	default: return T();
	}
}

void ConstantFolder::set_result(const Variant &value, bool literal)
{
	r_constant_value = value;
	r_constant = true;
	r_literal = literal;
}

void ConstantFolder::visit(RefPtr<Expression> &expr)
{
	r_constant_value = Variant();
	r_constant = false;
	r_literal = false;
	r_uses_iter_var = false;
	expr->visit(*this);
	/* Don't replace literals since they'd only be replaced with an identical
	literal.  Also skip anything that uses an iteration variable, but pass on
	the result so the Iteration visiting function can handle it. */
	if(!r_constant || r_literal || r_uses_iter_var)
		return;

	BasicTypeDeclaration::Kind kind = get_value_kind(r_constant_value);
	if(kind==BasicTypeDeclaration::VOID)
	{
		r_constant = false;
		return;
	}

	RefPtr<Literal> literal = new Literal;
	if(kind==BasicTypeDeclaration::BOOL)
		literal->token = (r_constant_value.value<bool>() ? "true" : "false");
	else if(kind==BasicTypeDeclaration::INT)
		literal->token = lexical_cast<string>(r_constant_value.value<int>());
	else if(kind==BasicTypeDeclaration::FLOAT)
		literal->token = lexical_cast<string>(r_constant_value.value<float>());
	literal->value = r_constant_value;
	expr = literal;
}

void ConstantFolder::visit(Literal &literal)
{
	set_result(literal.value, true);
}

void ConstantFolder::visit(VariableReference &var)
{
	/* If an iteration variable is initialized with a constant value, return
	that value here for the purpose of evaluating the loop condition for the
	first iteration. */
	if(var.declaration==iteration_var)
	{
		set_result(iter_init_value);
		r_uses_iter_var = true;
	}
}

void ConstantFolder::visit(MemberAccess &memacc)
{
	TraversingVisitor::visit(memacc);
	r_constant = false;
}

void ConstantFolder::visit(Swizzle &swizzle)
{
	TraversingVisitor::visit(swizzle);
	r_constant = false;
}

void ConstantFolder::visit(UnaryExpression &unary)
{
	TraversingVisitor::visit(unary);
	bool can_fold = r_constant;
	r_constant = false;
	if(!can_fold)
		return;

	BasicTypeDeclaration::Kind kind = get_value_kind(r_constant_value);

	char oper = unary.oper->token[0];
	char oper2 = unary.oper->token[1];
	if(oper=='!')
	{
		if(kind==BasicTypeDeclaration::BOOL)
			set_result(!r_constant_value.value<bool>());
	}
	else if(oper=='~')
	{
		if(kind==BasicTypeDeclaration::INT)
			set_result(~r_constant_value.value<int>());
	}
	else if(oper=='-' && !oper2)
	{
		if(kind==BasicTypeDeclaration::INT)
			set_result(-r_constant_value.value<int>());
		else if(kind==BasicTypeDeclaration::FLOAT)
			set_result(-r_constant_value.value<float>());
	}
}

void ConstantFolder::visit(BinaryExpression &binary)
{
	visit(binary.left);
	bool left_constant = r_constant;
	bool left_iter_var = r_uses_iter_var;
	Variant left_value = r_constant_value;
	visit(binary.right);
	if(left_iter_var)
		r_uses_iter_var = true;

	bool can_fold = (left_constant && r_constant);
	r_constant = false;
	if(!can_fold)
		return;

	BasicTypeDeclaration::Kind left_kind = get_value_kind(left_value);
	BasicTypeDeclaration::Kind right_kind = get_value_kind(r_constant_value);
	// Currently only expressions with both sides of equal types are handled.
	if(left_kind!=right_kind)
		return;

	char oper = binary.oper->token[0];
	char oper2 = binary.oper->token[1];
	if(oper=='&' || oper=='|' || oper=='^')
	{
		if(oper2==oper && left_kind==BasicTypeDeclaration::BOOL)
			set_result(evaluate_logical(oper, left_value.value<bool>(), r_constant_value.value<bool>()));
		else if(!oper2 && left_kind==BasicTypeDeclaration::INT)
			set_result(evaluate_logical(oper, left_value.value<int>(), r_constant_value.value<int>()));
	}
	else if((oper=='<' || oper=='>') && oper2!=oper)
	{
		if(left_kind==BasicTypeDeclaration::INT)
			set_result(evaluate_relation(binary.oper->token, left_value.value<int>(), r_constant_value.value<int>()));
		else if(left_kind==BasicTypeDeclaration::FLOAT)
			set_result(evaluate_relation(binary.oper->token, left_value.value<float>(), r_constant_value.value<float>()));
	}
	else if((oper=='=' || oper=='!') && oper2=='=')
	{
		if(left_kind==BasicTypeDeclaration::INT)
			set_result((left_value.value<int>()==r_constant_value.value<int>()) == (oper=='='));
		if(left_kind==BasicTypeDeclaration::FLOAT)
			set_result((left_value.value<float>()==r_constant_value.value<float>()) == (oper=='='));
	}
	else if(oper=='+' || oper=='-' || oper=='*' || oper=='/')
	{
		if(left_kind==BasicTypeDeclaration::INT)
			set_result(evaluate_arithmetic(oper, left_value.value<int>(), r_constant_value.value<int>()));
		else if(left_kind==BasicTypeDeclaration::FLOAT)
			set_result(evaluate_arithmetic(oper, left_value.value<float>(), r_constant_value.value<float>()));
	}
	else if(oper=='%' || ((oper=='<' || oper=='>') && oper2==oper))
	{
		if(left_kind!=BasicTypeDeclaration::INT)
			return;

		if(oper=='%')
			set_result(left_value.value<int>()%r_constant_value.value<int>());
		else if(oper=='<')
			set_result(left_value.value<int>()<<r_constant_value.value<int>());
		else if(oper=='>')
			set_result(left_value.value<int>()>>r_constant_value.value<int>());
	}
}

void ConstantFolder::visit(Assignment &assign)
{
	TraversingVisitor::visit(assign);
	r_constant = false;
}

void ConstantFolder::visit(TernaryExpression &ternary)
{
	TraversingVisitor::visit(ternary);
	r_constant = false;
}

void ConstantFolder::visit(FunctionCall &call)
{
	TraversingVisitor::visit(call);
	r_constant = false;
}

void ConstantFolder::visit(VariableDeclaration &var)
{
	if(iteration_init && var.init_expression)
	{
		visit(var.init_expression);
		if(r_constant)
		{
			/* Record the value of a constant initialization expression of an
			iteration, so it can be used to evaluate the loop condition. */
			iteration_var = &var;
			iter_init_value = r_constant_value;
		}
	}
	else
		TraversingVisitor::visit(var);
}

void ConstantFolder::visit(Iteration &iter)
{
	SetForScope<Block *> set_block(current_block, &iter.body);

	/* The iteration variable is not normally inlined into expressions, so we
	process it specially here.  If the initial value causes the loop condition
	to evaluate to false, then the expression can be folded. */
	iteration_var = 0;
	if(iter.init_statement)
	{
		SetFlag set_init(iteration_init);
		iter.init_statement->visit(*this);
	}

	if(iter.condition)
	{
		visit(iter.condition);
		if(r_constant && r_constant_value.check_type<bool>() && !r_constant_value.value<bool>())
		{
			RefPtr<Literal> literal = new Literal;
			literal->token = "false";
			literal->value = r_constant_value;
			iter.condition = literal;
		}
	}
	iteration_var = 0;

	iter.body.visit(*this);
	if(iter.loop_expression)
		visit(iter.loop_expression);
}


void ConstantConditionEliminator::apply(Stage &stage)
{
	stage.content.visit(*this);
	NodeRemover().apply(stage, nodes_to_remove);
}

ConstantConditionEliminator::ConstantStatus ConstantConditionEliminator::check_constant_condition(const Expression &expr)
{
	if(const Literal *literal = dynamic_cast<const Literal *>(&expr))
		if(literal->value.check_type<bool>())
			return (literal->value.value<bool>() ? CONSTANT_TRUE : CONSTANT_FALSE);
	return NOT_CONSTANT;
}

void ConstantConditionEliminator::visit(Block &block)
{
	SetForScope<Block *> set_block(current_block, &block);
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		insert_point = i;
		(*i)->visit(*this);
	}
}

void ConstantConditionEliminator::visit(RefPtr<Expression> &expr)
{
	r_ternary_result = 0;
	expr->visit(*this);
	if(r_ternary_result)
		expr = r_ternary_result;
	r_ternary_result = 0;
}

void ConstantConditionEliminator::visit(TernaryExpression &ternary)
{
	ConstantStatus result = check_constant_condition(*ternary.condition);
	if(result!=NOT_CONSTANT)
		r_ternary_result = (result==CONSTANT_TRUE ? ternary.true_expr : ternary.false_expr);
	else
		r_ternary_result = 0;
}

void ConstantConditionEliminator::visit(Conditional &cond)
{
	ConstantStatus result = check_constant_condition(*cond.condition);
	if(result!=NOT_CONSTANT)
	{
		Block &block = (result==CONSTANT_TRUE ? cond.body : cond.else_body);
		// TODO should check variable names for conflicts.  Potentially reuse InlineContentInjector?
		current_block->body.splice(insert_point, block.body);
		nodes_to_remove.insert(&cond);
		return;
	}

	TraversingVisitor::visit(cond);
}

void ConstantConditionEliminator::visit(Iteration &iter)
{
	if(iter.condition)
	{
		ConstantStatus result = check_constant_condition(*iter.condition);
		if(result==CONSTANT_FALSE)
		{
			nodes_to_remove.insert(&iter);
			return;
		}
	}

	TraversingVisitor::visit(iter);
}


bool UnusedTypeRemover::apply(Stage &stage)
{
	stage.content.visit(*this);
	NodeRemover().apply(stage, unused_nodes);
	return !unused_nodes.empty();
}

void UnusedTypeRemover::visit(Literal &literal)
{
	unused_nodes.erase(literal.type);
}

void UnusedTypeRemover::visit(UnaryExpression &unary)
{
	unused_nodes.erase(unary.type);
	TraversingVisitor::visit(unary);
}

void UnusedTypeRemover::visit(BinaryExpression &binary)
{
	unused_nodes.erase(binary.type);
	TraversingVisitor::visit(binary);
}

void UnusedTypeRemover::visit(TernaryExpression &ternary)
{
	unused_nodes.erase(ternary.type);
	TraversingVisitor::visit(ternary);
}

void UnusedTypeRemover::visit(FunctionCall &call)
{
	unused_nodes.erase(call.type);
	TraversingVisitor::visit(call);
}

void UnusedTypeRemover::visit(BasicTypeDeclaration &type)
{
	if(type.base_type)
		unused_nodes.erase(type.base_type);
	unused_nodes.insert(&type);
}

void UnusedTypeRemover::visit(ImageTypeDeclaration &type)
{
	if(type.base_type)
		unused_nodes.erase(type.base_type);
	unused_nodes.insert(&type);
}

void UnusedTypeRemover::visit(StructDeclaration &strct)
{
	unused_nodes.insert(&strct);
	TraversingVisitor::visit(strct);
}

void UnusedTypeRemover::visit(VariableDeclaration &var)
{
	unused_nodes.erase(var.type_declaration);
}

void UnusedTypeRemover::visit(InterfaceBlock &iface)
{
	unused_nodes.erase(iface.type_declaration);
}

void UnusedTypeRemover::visit(FunctionDeclaration &func)
{
	unused_nodes.erase(func.return_type_declaration);
	TraversingVisitor::visit(func);
}


UnusedVariableRemover::UnusedVariableRemover():
	stage(0),
	interface_block(0),
	r_assignment(0),
	assignment_target(false),
	r_side_effects(false)
{ }

bool UnusedVariableRemover::apply(Stage &s)
{
	stage = &s;
	s.content.visit(*this);

	for(list<AssignmentInfo>::const_iterator i=assignments.begin(); i!=assignments.end(); ++i)
		if(i->used_by.empty())
			unused_nodes.insert(i->node);

	for(map<string, InterfaceBlock *>::const_iterator i=s.interface_blocks.begin(); i!=s.interface_blocks.end(); ++i)
		if(i->second->instance_name.empty())
			unused_nodes.insert(i->second);

	for(BlockVariableMap::const_iterator i=variables.begin(); i!=variables.end(); ++i)
	{
		if(i->second.output)
		{
			/* The last visible assignments of output variables are used by the
			next stage or the API. */
			for(vector<AssignmentInfo *>::const_iterator j=i->second.assignments.begin(); j!=i->second.assignments.end(); ++j)
				unused_nodes.erase((*j)->node);
		}

		if(!i->second.output && !i->second.referenced)
		{
			// Don't remove variables from inside interface blocks.
			if(!i->second.interface_block)
				unused_nodes.insert(i->first);
		}
		else if(i->second.interface_block)
			// Interface blocks are kept if even one member is used.
			unused_nodes.erase(i->second.interface_block);
	}

	NodeRemover().apply(s, unused_nodes);

	return !unused_nodes.empty();
}

void UnusedVariableRemover::referenced(const Assignment::Target &target, Node &node)
{
	VariableInfo &var_info = variables[target.declaration];
	var_info.referenced = true;
	if(!assignment_target)
	{
		for(vector<AssignmentInfo *>::const_iterator i=var_info.assignments.begin(); i!=var_info.assignments.end(); ++i)
			(*i)->used_by.push_back(&node);
	}
}

void UnusedVariableRemover::visit(VariableReference &var)
{
	referenced(var.declaration, var);
}

void UnusedVariableRemover::visit(InterfaceBlockReference &iface)
{
	referenced(iface.declaration, iface);
}

void UnusedVariableRemover::visit(UnaryExpression &unary)
{
	TraversingVisitor::visit(unary);
	if(unary.oper->token[1]=='+' || unary.oper->token[1]=='-')
		r_side_effects = true;
}

void UnusedVariableRemover::visit(BinaryExpression &binary)
{
	if(binary.oper->token[0]=='[')
	{
		binary.left->visit(*this);
		SetFlag set(assignment_target, false);
		binary.right->visit(*this);
	}
	else
		TraversingVisitor::visit(binary);
}

void UnusedVariableRemover::visit(Assignment &assign)
{
	{
		SetFlag set(assignment_target, (assign.oper->token[0]=='='));
		assign.left->visit(*this);
	}
	assign.right->visit(*this);
	r_assignment = &assign;
	r_side_effects = true;
}

void UnusedVariableRemover::visit(FunctionCall &call)
{
	TraversingVisitor::visit(call);
	/* Treat function calls as having side effects so expression statements
	consisting of nothing but a function call won't be optimized away. */
	r_side_effects = true;

	if(stage->type==Stage::GEOMETRY && call.name=="EmitVertex")
	{
		for(map<Statement *, VariableInfo>::const_iterator i=variables.begin(); i!=variables.end(); ++i)
			if(i->second.output)
				referenced(i->first, call);
	}
}

void UnusedVariableRemover::record_assignment(const Assignment::Target &target, Node &node)
{
	assignments.push_back(AssignmentInfo());
	AssignmentInfo &assign_info = assignments.back();
	assign_info.node = &node;
	assign_info.target = target;

	/* An assignment to the target hides any assignments to the same target or
	its subfields. */
	VariableInfo &var_info = variables[target.declaration];
	for(unsigned i=0; i<var_info.assignments.size(); ++i)
	{
		const Assignment::Target &t = var_info.assignments[i]->target;

		bool subfield = (t.chain_len>=target.chain_len);
		for(unsigned j=0; (subfield && j<target.chain_len); ++j)
			subfield = (t.chain[j]==target.chain[j]);

		if(subfield)
			var_info.assignments.erase(var_info.assignments.begin()+i);
		else
			++i;
	}

	var_info.assignments.push_back(&assign_info);
}

void UnusedVariableRemover::visit(ExpressionStatement &expr)
{
	r_assignment = 0;
	r_side_effects = false;
	TraversingVisitor::visit(expr);
	if(r_assignment && r_assignment->target.declaration)
		record_assignment(r_assignment->target, expr);
	if(!r_side_effects)
		unused_nodes.insert(&expr);
}

void UnusedVariableRemover::visit(VariableDeclaration &var)
{
	VariableInfo &var_info = variables[&var];
	var_info.interface_block = interface_block;

	/* Mark variables as output if they're used by the next stage or the
	graphics API. */
	if(interface_block)
		var_info.output = (interface_block->interface=="out" && (interface_block->linked_block || !interface_block->block_name.compare(0, 3, "gl_")));
	else
		var_info.output = (var.interface=="out" && (stage->type==Stage::FRAGMENT || var.linked_declaration || !var.name.compare(0, 3, "gl_")));

	if(var.init_expression)
	{
		var_info.initialized = true;
		record_assignment(&var, *var.init_expression);
	}
	TraversingVisitor::visit(var);
}

void UnusedVariableRemover::visit(InterfaceBlock &iface)
{
	if(iface.instance_name.empty())
	{
		SetForScope<InterfaceBlock *> set_block(interface_block, &iface);
		iface.struct_declaration->members.visit(*this);
	}
	else
	{
		VariableInfo &var_info = variables[&iface];
		var_info.output = (iface.interface=="out" && (iface.linked_block || !iface.block_name.compare(0, 3, "gl_")));
	}
}

void UnusedVariableRemover::merge_variables(const BlockVariableMap &other_vars)
{
	for(BlockVariableMap::const_iterator i=other_vars.begin(); i!=other_vars.end(); ++i)
	{
		BlockVariableMap::iterator j = variables.find(i->first);
		if(j!=variables.end())
		{
			/* The merged blocks started as copies of each other so any common
			assignments must be in the beginning. */
			unsigned k = 0;
			for(; (k<i->second.assignments.size() && k<j->second.assignments.size()); ++k)
				if(i->second.assignments[k]!=j->second.assignments[k])
					break;

			// Remaining assignments are unique to each block; merge them.
			j->second.assignments.insert(j->second.assignments.end(), i->second.assignments.begin()+k, i->second.assignments.end());
			j->second.referenced |= i->second.referenced;
		}
		else
			variables.insert(*i);
	}
}

void UnusedVariableRemover::visit(FunctionDeclaration &func)
{
	if(func.body.body.empty())
		return;

	BlockVariableMap saved_vars = variables;
	// Assignments from other functions should not be visible.
	for(BlockVariableMap::iterator i=variables.begin(); i!=variables.end(); ++i)
		i->second.assignments.resize(i->second.initialized);
	TraversingVisitor::visit(func);
	swap(variables, saved_vars);
	merge_variables(saved_vars);

	/* Always treat function parameters as referenced.  Removing unused
	parameters is not currently supported. */
	for(NodeArray<VariableDeclaration>::iterator i=func.parameters.begin(); i!=func.parameters.end(); ++i)
	{
		BlockVariableMap::iterator j = variables.find(i->get());
		if(j!=variables.end())
			j->second.referenced = true;
	}
}

void UnusedVariableRemover::visit(Conditional &cond)
{
	cond.condition->visit(*this);
	BlockVariableMap saved_vars = variables;
	cond.body.visit(*this);
	swap(saved_vars, variables);
	cond.else_body.visit(*this);

	/* Visible assignments after the conditional is the union of those visible
	at the end of the if and else blocks.  If there was no else block, then it's
	the union of the if block and the state before it. */
	merge_variables(saved_vars);
}

void UnusedVariableRemover::visit(Iteration &iter)
{
	BlockVariableMap saved_vars = variables;
	TraversingVisitor::visit(iter);

	/* Merge assignments from the iteration, without clearing previous state.
	Further analysis is needed to determine which parts of the iteration body
	are always executed, if any. */
	merge_variables(saved_vars);
}


bool UnusedFunctionRemover::apply(Stage &stage)
{
	stage.content.visit(*this);
	NodeRemover().apply(stage, unused_nodes);
	return !unused_nodes.empty();
}

void UnusedFunctionRemover::visit(FunctionCall &call)
{
	TraversingVisitor::visit(call);

	unused_nodes.erase(call.declaration);
	if(call.declaration && call.declaration->definition!=call.declaration)
		used_definitions.insert(call.declaration->definition);
}

void UnusedFunctionRemover::visit(FunctionDeclaration &func)
{
	TraversingVisitor::visit(func);

	if((func.name!="main" || func.body.body.empty()) && !used_definitions.count(&func))
		unused_nodes.insert(&func);
}

} // namespace SL
} // namespace GL
} // namespace Msp
